from moontour_common.models.coordinates import Coordinates
from moontour_common.models.guess import Guess
from moontour_common.models.phase import PhaseStatus, Phase
from moontour_common.models.rooms.duel_room import DuelPlayer, DuelRoom
from moontour_common.models.rooms.room import RoomStatus, BaseRoom
from moontour_common.models.rooms.streak_room import StreakRoom
from moontour_common.models.rooms.teams_room import Team, TeamsRoom
from moontour_common.models.user import User, UserCreate, InternalUser



from pydantic import BaseModel
from pydantic.color import Color

from moontour_common.models import Guess
from moontour_common.models.rooms.health_room import HealthRoom, START_HEALTH, HealthPhase
from moontour_common.models.rooms.room import RoomMode
from moontour_common.models.user import User


class Team(BaseModel):
    color: Color
    health: int = START_HEALTH
    players: list[User] = []


class TeamsPhase(HealthPhase):
    guesses: dict[str, Guess] = {}  # User ID to guess


class TeamsRoom(HealthRoom):
    @staticmethod
    def get_mode() -> RoomMode:
        return RoomMode.teams

    team_size: int = 2
    teams: dict[str, Team] = {
        'red': Team(color=Color('red')),
        'blue': Team(color=Color('blue')),
    }
    phases: list[TeamsPhase] = []



import json
import os
from contextlib import asynccontextmanager
from typing import TypeVar, AsyncContextManager

import redis.asyncio as redis
from redis.commands.json.path import Path

from moontour_common.database.rabbitmq import notify_room
from moontour_common.models import BaseRoom

_RoomType = TypeVar('_RoomType', bound=BaseRoom)

ROOM_KEY_PREFIX = 'room:'

_host = os.getenv('REDIS_HOST', 'redis')
redis_client = redis.Redis(host=_host)


def get_room_key(room_id: str) -> str:
    return f'{ROOM_KEY_PREFIX}{room_id}'


@asynccontextmanager
async def room_lock(room_id: str):
    async with redis_client.lock(f'room-{room_id}'):
        yield


async def create_room(room: _RoomType):
    await set_room(room)


def delete_room(room_id: str):
    redis_client.json().delete(get_room_key(room_id), Path.root_path())


async def get_room(room_id: str, model: type[_RoomType]) -> _RoomType:
    room_raw = await redis_client.execute_command('JSON.GET', get_room_key(room_id), Path.root_path())
    return model.parse_raw(room_raw)


async def set_room(room: _RoomType):
    await redis_client.execute_command('JSON.SET', get_room_key(room.id), Path.root_path(), room.json())


@asynccontextmanager
async def modify_room(
        room_id: str,
        model: type[_RoomType] = BaseRoom,
        notify: bool = True
) -> AsyncContextManager[_RoomType]:
    assert model.get_mode() is not None  # Modifying abstract models will lead to unexpected results

    async with room_lock(room_id):
        room = await get_room(room_id, model)
        try:
            yield room
        except Exception:
            raise
        else:
            await set_room(room)

            if notify:
                await notify_room(room)



from datetime import datetime
from enum import StrEnum

from pydantic import BaseModel, Field

from moontour_common.models.coordinates import Coordinates


class PhaseStatus(StrEnum):
    waiting = 'waiting'
    running = 'running'
    finished = 'finished'


class Phase(BaseModel):
    target: Coordinates
    status: PhaseStatus = PhaseStatus.waiting
    create_time: datetime = Field(default_factory=datetime.now)
    start_time: datetime | None = None
    end_time: datetime | None



import uuid
from datetime import datetime
from enum import StrEnum

from pydantic import BaseModel, Field, validator


class RoomMode(StrEnum):
    streak = 'streak'
    duel = 'duel'
    teams = 'teams'


class RoomStatus(StrEnum):
    waiting = 'waiting'
    waiting_to_start = 'waitingToStart'
    running = 'running'
    closed = 'closed'


class BaseRoom(BaseModel):
    id: str = Field(alias='id_', default_factory=lambda: str(uuid.uuid4()))
    map: str = 'world'
    status: RoomStatus = RoomStatus.waiting
    create_time: datetime = Field(default_factory=datetime.now)
    start_time: datetime | None = None
    mode: RoomMode | None = None

    @validator("mode", always=True)
    def compute_mode(cls, value):
        return cls.get_mode() or value

    @staticmethod
    def get_mode() -> RoomMode | None:
        return None

    class Config:
        allow_population_by_field_name = True



from moontour_common.models import Phase
from moontour_common.models.rooms.room import BaseRoom

START_HEALTH = 5000


class HealthRoom(BaseRoom):
    start_health = START_HEALTH
    guess_duration: float = 15  # Time between first guess to phase ending


class HealthPhase(Phase):
    damage_multiplier: float = 1



from moontour_common.models import Guess
from moontour_common.models.rooms.health_room import HealthRoom, START_HEALTH, HealthPhase
from moontour_common.models.rooms.room import RoomMode
from moontour_common.models.user import User


class DuelPlayer(User):
    health: int = START_HEALTH


class DuelPhase(HealthPhase):
    guesses: dict[str, Guess] = {}  # User ID to guess


class DuelRoom(HealthRoom):
    @staticmethod
    def get_mode() -> RoomMode:
        return RoomMode.duel

    player_count: int = 2
    players: list[DuelPlayer] = []
    phases: list[DuelPhase] = []



from datetime import datetime

import geopy.distance
from pydantic import BaseModel, Field, validator

from moontour_common.models import Coordinates

MAX_POINTS = 5000


class Guess(BaseModel):
    time: datetime = Field(default_factory=datetime.now)
    coordinates: Coordinates
    target: Coordinates
    distance: float | None
    points: int | None

    @validator('distance', always=True)
    def compute_distance(cls, v, values):
        guess = values['coordinates']
        target = values['target']
        return geopy.distance.geodesic(
            (target.latitude, target.longitude),
            (guess.latitude, guess.longitude)
        ).km

    @validator('points', always=True)
    def compute_points(cls, v, values):
        distance = values['distance']
        if distance is None:
            return None
        return max(MAX_POINTS - int(distance), 0)



from moontour_common.models.guess import Guess
from moontour_common.models.phase import Phase
from moontour_common.models.rooms.room import BaseRoom, RoomMode
from moontour_common.models.user import User


class StreaksPhase(Phase):
    guess: Guess | None = None


class StreakRoom(BaseRoom):
    @staticmethod
    def get_mode() -> RoomMode:
        return RoomMode.streak

    streak: int = 0
    phases: list[StreaksPhase] = []
    player: User | None



import os

import aio_pika
from aio_pika import Message
from aio_pika.abc import AbstractChannel, AbstractExchange, AbstractQueue, ExchangeType

from moontour_common.models import BaseRoom

RABBITMQ_HOST = os.getenv('RABBITMQ_HOST', 'rabbitmq')
ROOMS_EXCHANGE_NAME = 'rooms'

_channel: AbstractChannel | None = None


async def get_rabbitmq_client() -> AbstractChannel:
    global _channel
    if _channel is None:
        connection = await aio_pika.connect_robust(host=RABBITMQ_HOST)
        _channel = await connection.channel()
    return _channel


async def get_rooms_exchange() -> AbstractExchange:
    client = await get_rabbitmq_client()
    return await client.declare_exchange(name=ROOMS_EXCHANGE_NAME, type=ExchangeType.TOPIC)


async def declare_player_queue(room_id: str, user_id: str) -> AbstractQueue:
    client = await get_rabbitmq_client()
    queue_name = f'{room_id}.{user_id}'
    queue = await client.declare_queue(name=queue_name, auto_delete=True)
    exchange = await get_rooms_exchange()
    await queue.bind(exchange, routing_key=f'{room_id}.*')
    return queue


async def notify_room(room: BaseRoom):
    exchange = await get_rooms_exchange()
    await exchange.publish(
        message=Message(body=room.json().encode()),
        routing_key=f'{room.id}.*',
    )
